#ifndef POLYHEDRON_H
#define POLYHEDRON_H

#include "realnumber.h"
#include "vectormatrix.h"
#include <vector> //For "vector" template used for vectors, which are arrays
using namespace std; //Need this for standard template library to work.

//Helper methods
void setMaterialColor(cVector3 color); 
void setMaterialColor(Real r, Real g, Real b, Real alpha = 1.0);
void setVertexColor(cVector3 color, float alpha = 1.0);
void glutWireCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks);
void glutSolidCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks);

class cAssembly
{
public:
	static int MAXDEPTH; /* Try 3 for this default value to control how deep
		down into composite levels to go. */
		/* We have some bit flags that we can OR togetehr to get drawflag
		arguments to our draw(int drawflag) methods. */
	static const int WIREFRAMEFLAG; //Draw using wireframe only.
	static const int SHOWCONNECTORFLAG;  /* Draw lines between the _position
		of each cAssemblyComposite and _position of each of its child assemblies. */
	static const int SHOWRADIUSFLAG;/* If drawflags inscludes SHOWRADIUSFLAG, 
		I draw a wireframe sphere of _actualradius at the same spot as
		 each assembly to check if it has the right value. */

	static cAssembly* cAssembly::makeRandomAssembly(int maxdepth = MAXDEPTH); 
		/* makeRandomAssembly is a static (think "global") method that creates
		random assembliess. Methods of this kind are called factory methods,
		in that they create new objects for you.  The meaning of the maxdepth
		argument is that we guarantee not to return a nested cAssemblyComposite
		with more than maxdepth levels.  You use this method in your main program
		by picking some value of MAXDEPTH maybe between 2 and 5 and writing a
		line like
		cAssembly *pnew = cAssembly::makeRandomAssembly(MAXDEPTH); 		 */
protected:
//Geometry
	Real _actualradius; /* Make this be the actual radius of the most
		distant point form the origin. */
//Transformation
	cVector _position;
	cVector _scale;
	cMatrix _rotationmatrix; /* We use _rotationmatrix to accumulate our
		succesive rotations. During the anaimate(dt) step we compute a
		change cMatrix drotation and multiply it on the RIGHT of our
		_rotationmatrix, which means that the additional animation rotation
		acts on the vertex BEFORE the old _rotationmatrix.  This way any
		initial adjustments of the assembly's rotation still move
		the assembly as before. */
//Motion
	Real _animatespinrate; /* animate(dt) normally applies a rotation of dt*_animatespinrate
		around the _animatespinaxis to _rotationmatrix at each call. */
	cVector _animatespinaxis; /*_animatespinaxis doesn't reflect the current
		state of _rotationmatrix.  It's used in the animate method as the
		axis to spin the object around, but _rotationmatrix may contain
		other rotations. */
	Real _cumulativeanimatespinangle; /* _cumulativeanimatespinangle is a dummy, tracking variable.  It 
		measures the amount of dt*_animatespinrate that has acumulated in the
		animate method.  The real rotation lives inside _rotationmatrix. */
	cVector _velocity;
public:
//Construtor, Destructor, Copy
	cAssembly();
	cAssembly(const cAssembly &assembly); 
	cAssembly(cAssembly *passembly); 
	virtual ~cAssembly(){} // C++ parent class destructors should always be virtual if y
	void copy(const cAssembly &assembly); 
//Helper methods
	virtual void fixActualRadius(){} 
//Mutators
	//Randomizer
	virtual void randomize();
		//Randomizes _animatespinaxis, _animatespinrate, _rotationmatrix
	//Field Mutators
	//Absolute mutators
	void setPosition(const cVector &newposition){_position = newposition;}
	void setPosition(Real px, Real py, Real pz)
		{setPosition(cVector(px, py, pz));}
	void setScale(Real sx, Real sy, Real sz){_scale = cVector(sx,sy,sz);} 
	void setScale(Real s){_scale = cVector(s, s, s);}
	void setRotation(Real rdegrees, Real rx, Real ry, Real rz); /* Sets
		_rotationmatrix to be a rotation of rdegrees around the
		axis (rx, ry, rz). */
	void setRotation(Real redgrees, cVector saxis)
		{setRotation(redgrees, saxis.x(), saxis.y(), saxis.z());}
	void setAnimateSpinAxis(cVector saxis){_animatespinaxis = saxis;}
	void setAnimateSpinAxis(Real spx, Real spy, Real spz)
		{setAnimateSpinAxis(cVector(spx, spy, spz));}
   	void setAnimateSpinRate(Real sangle){_animatespinrate = sangle;}
	//Reset mutators
	void clearRotation(){setRotation(0.0, 0.0, 0.0, 1.0);} /* Resets 
		_rotationmatrix to a default do-nothing value. */
	void clearScale(){setScale(1.0);}
	void clearPosition(){setPosition(cVector(0.0, 0.0, 0.0));}
	//Relative mutators
	void applyRotation(Real rdegrees, Real rx, Real ry, Real rz);
		/* Right-multiplies the current _rotationmatrix by a rotation of
		rdegrees around	the axis (rx, ry, rz).  The applied rotation will
		happen BEFORE the rotation that was already inside _rotationmatrix. */
	void applyRotation(Real redgrees, cVector drotation)
		{applyRotation(redgrees, drotation.x(), drotation.y(), drotation.z());}
	void multiplyScale(cVector scalevector); /* Use multiplyScale if you want
		to accumulate scale changes. */
	void multiplyScale(Real s);/* Use this if you want to accumulate
		scale changes. */
	void translate(cVector dposition){_position += dposition;}
//Accessors
	Real maxscale(){return __max(_scale.x(), __max(_scale.y(), _scale.z()));}
	//Geometry accessors
	Real radius(){ return _actualradius * maxscale();}
	cVector position()const{return _position;}
	cMatrix rotationMatrix()const{return _rotationmatrix;}
	//Complexity accessor
	virtual int depth(){return 0;} //Measures how many nestings it has.
//Essential methods
	virtual void animate(Real dt);
	virtual void draw(int drawflags = 0){}
		//If drawflags is 1, we draw wireframe.
};

class cAssemblyComposite : public cAssembly
{
private:
	vector <cAssembly*> _pchild;
public:
//Constructor, etc.
	cAssemblyComposite(){} //Default constructor makes empty _pchild array
	cAssemblyComposite(const cAssemblyComposite &comp); //Uses copy
	cAssemblyComposite(cAssemblyComposite *pcomp); //Uses copy
	virtual ~cAssemblyComposite(); 
		//Destructor has to delete all the _pchild[i].
	void copy(const cAssemblyComposite &comp); //Need to handle _pchild.
	cAssemblyComposite operator=(const cAssemblyComposite &comp);
		//uses the copy method
//Special methods
	void add(cAssembly *pa);
	void addRandomChildren(int maxdepth = cAssembly::MAXDEPTH);
//Overloads
	virtual void fixActualRadius(); /* Try to adjust 
		 _actualradius to include all the pieces. */
	virtual int depth(); //Return 1 + max depth of children.
	virtual void animate(Real dt);
	virtual void draw(int drawflags = 0);
};

class cPolyhedron : public cAssembly
{
public:
//const statics 
	static const int TETRAHEDRON;
	static const int CUBE;
	static const int OCTAHEDRON;
	static const int ICOSOHEDRON;
	static const int DODECAHEDRON;
	static const int SPHERE;
	static const int CYLINDER;
	static const int CONE;
	static const int TORUS;
	static const int TEAPOT;
protected:
//Type
	int _gluttype;
//Polyhedron Geometry
	Real _radius; //For SPHERE, TORUS
	int _slices; //For SPHERE, TORUS, CYLINDER, CONE
	int _stacks; //For SPHERE, TORUS, CYLINDER, CONE
	Real _edge; //For CUBE
	Real _tuberadius; //For TORUS
	Real _bottomradius; //For CYLINDER and CONE
	Real _topradius; //For CYLINDER
	Real _height; //For CONE, CYLINDER
//Appearance
	cVector3 _color;
	Real _alpha;
	bool _filled;
	Real _linewidth;
public:
//Constructor, destructor, operator=
	cPolyhedron(int gluttype = SPHERE, Real radius = 1.0,
		int slices = 32, int stacks = 16, Real edge = sqrt(4.0/3.0),
		Real tuberadius = 0.3,	Real bottomradius = 1.0,
		Real topradius = 1.0 , Real height = 3.0);
		/* I select the default edge value to make a cube of radius 1. */
	cPolyhedron(const cPolyhedron &poly); //Uses default operator=
	cPolyhedron(cPolyhedron *ppoly); //Uses default operator=
	virtual	~cPolyhedron(){}; /* The destructor of a base class must
		be virtual if any of its child classes can have more fields that
		may need additional	destruction. */
	void copy(const cPolyhedron &poly);
//Helper functions
	virtual void fixActualRadius(); /* When you construct a polyhedron you
		may need to fix the _scale to make it truly be the size that
		you expect.  In particular,	we want to scale a default dodecahedron
		so it has unit radius like the other Platonic solids
		(except for cube, which takes an edge argument to its constructor). */
//Mutators
	//Color and draw style Mutators
	void setFilled(bool yesno){_filled = yesno;}
	void setColor(cVector3 col){_color = col;}
	void setLineWidth(Real linewidth){_linewidth = linewidth;}
	void setAlpha(Real alpha){_alpha = alpha;}
	//Shape Mutators
	void setGluttype(int gluttype){_gluttype = gluttype;}
	void setSlices(int slices){_slices = slices;}
	void setStacks(int stacks){_stacks = stacks;}
	void setEdge(Real edge){_edge = edge; fixActualRadius();}
	void setTubeRadius(Real tuberadius){_tuberadius = tuberadius;}
	void setBottomRadius(Real bottomradius)
		{_bottomradius = bottomradius; fixActualRadius();}
	void setTopRadius(Real topradius)
		{_topradius = topradius; fixActualRadius();}
	void setHeight(Real height){_height = height; fixActualRadius();}
//Accessors
	//Color and style accessors
	cVector color() const{return _color;}
	Real alpha()const{return _alpha;}
	bool filled() const{return _filled;}
	Real lineWidth() const {return _linewidth;}
	//Shape accessors
	int glutType(){return _gluttype;}
//Overloads
	virtual int depth(){return 0;}
	virtual void randomize(); 
		/* Calls cAssembly::randomize.  Randomizes animatespinaxis, 
		animatespinrate, animatespinrate, 
		_rotationmatrix, shape, color, fill. */
	virtual void draw(int drawflags = 0); 
		//If drawflags is 1, we draw wireframe.
};
#endif //POLYHEDRON_H